from datetime import timedelta

from uuid import uuid4

from django.utils import timezone

from germanium.decorators import data_consumer
from germanium.test_cases.default import GermaniumTestCase
from germanium.tools import assert_equal, assert_true, assert_false, assert_raises

from auth_token.models import AuthorizationToken, MobileDevice, MobileDeviceAlreadyExists
from auth_token.config import settings

from .base import BaseTestCaseMixin


__all__ = (
    'TokenTestCase',
)


class TokenTestCase(BaseTestCaseMixin, GermaniumTestCase):

    @data_consumer('create_user')
    def test_should_return_proper_time_to_expiration(self, user):
        expired_token = AuthorizationToken.objects.create(
            user=user, ip='127.0.0.1', backend='test', expires_at=timezone.now()
        )
        expired_token = AuthorizationToken.objects.get(pk=expired_token.pk)
        assert_equal(AuthorizationToken.objects.get(pk=expired_token.pk).time_to_expiration, timedelta(seconds=0))

        non_expired_token = AuthorizationToken.objects.create(user=user, ip='127.0.0.1', backend='test')
        assert_true(non_expired_token.time_to_expiration.total_seconds() > 0)

    @data_consumer('create_user')
    def test_only_one_mobile_device_should_be_primary(self, user):
        mobile_device1 = MobileDevice.objects.activate_or_create(uuid4(), user, is_primary=True)
        mobile_device2 = MobileDevice.objects.activate_or_create(uuid4(), user, is_primary=True)

        assert_false(mobile_device1.refresh_from_db().is_primary)
        assert_true(mobile_device2.refresh_from_db().is_primary)

    def test_only_both_mobile_devices_with_different_users_should_be_primary(self):
        mobile_device1 = MobileDevice.objects.activate_or_create(uuid4(), self.create_user(username='test1'),
                                                                 is_primary=True)
        mobile_device2 = MobileDevice.objects.activate_or_create(uuid4(), self.create_user(username='test2'),
                                                                 is_primary=True)

        assert_true(mobile_device1.refresh_from_db().is_primary)
        assert_true(mobile_device2.refresh_from_db().is_primary)

    @data_consumer('create_user')
    def test_two_active_mobile_devices_with_same_uuid_and_user_should_raise_exception(self, user):
        uuid = uuid4()
        MobileDevice.objects.activate_or_create(uuid, user, is_primary=True)
        with assert_raises(MobileDeviceAlreadyExists):
            MobileDevice.objects.activate_or_create(uuid, user, is_primary=True)

    def test_two_active_mobile_devices_with_same_uuid_should_be_created(self):
        uuid = uuid4
        MobileDevice.objects.activate_or_create(uuid4, self.create_user(username='test1'), is_primary=True)
        MobileDevice.objects.activate_or_create(uuid4, self.create_user(username='test2'), is_primary=True)

    @data_consumer('create_user')
    def test_deactivated_mobile_device_should_be_recreated(self, user):
        uuid = uuid4
        mobile_device = MobileDevice.objects.activate_or_create(uuid, user, is_primary=True)
        mobile_device.change_and_save(is_active=False)
        mobile_device = MobileDevice.objects.activate_or_create(uuid, user, is_primary=False, name='test', slug='test')
        assert_true(mobile_device.is_active)
        assert_false(mobile_device.is_primary)
        assert_equal(mobile_device.name, 'test')
        assert_equal(mobile_device.slug, 'test')



from datetime import timedelta

from freezegun import freeze_time

from django.core.management import call_command
from django.test import override_settings
from django.utils import timezone

from germanium.decorators import data_consumer
from germanium.test_cases.default import GermaniumTestCase
from germanium.tools import assert_equal, test_call_command

from auth_token.models import AuthorizationToken, OneTimePassword, AuthorizationRequest
from auth_token.config import settings
from auth_token.utils import generate_key

from .base import BaseTestCaseMixin


__all__ = (
   'CleanTokensCommandTestCase',
)


class CleanTokensCommandTestCase(BaseTestCaseMixin, GermaniumTestCase):

    @data_consumer('create_user')
    def test_clean_authorization_tokens_should_remove_only_old_tokens(self, user):
        expired_tokens = [
            AuthorizationToken.objects.create(user=user, ip='127.0.0.1', backend='test') for _ in range(10)
        ]
        not_expired_tokens = [
            AuthorizationToken.objects.create(user=user, ip='127.0.0.1', backend='test', expires_at=timezone.now())
            for _ in range(settings.COUNT_USER_PRESERVED_TOKENS - 5)
        ]
        test_call_command('clean_authorization_tokens')
        assert_equal(AuthorizationToken.objects.filter(pk__in=[token.pk for token in not_expired_tokens]).count(),
                     settings.COUNT_USER_PRESERVED_TOKENS - 5)
        assert_equal(AuthorizationToken.objects.filter(pk__in=[token.pk for token in expired_tokens]).count(), 5)

    @override_settings(AUTH_TOKEN_OTP_EXPIRATION_RETENTION_PERIOD=60)
    @data_consumer('create_user')
    def test_clean_one_time_password_should_remove_only_inactive_or_expired_otp(self, user):
        old_expired_otp = [
            OneTimePassword.objects.create(
                key_generator=generate_key,
                expires_at=timezone.now() - timedelta(seconds=60),
                slug='test',
            ) for _ in range(10)
        ]
        new_expired_otp = [
            OneTimePassword.objects.create(
                key_generator=generate_key,
                expires_at=timezone.now() - timedelta(seconds=59),
                slug='test',
            ) for _ in range(10)
        ]
        inactive_otp = [
            OneTimePassword.objects.create(key_generator=generate_key, is_active=False, slug='test')
            for _ in range(10)
        ]
        not_expired_otp = [
            OneTimePassword.objects.create(key_generator=generate_key, slug='test')
            for _ in range(10)
        ]
        test_call_command('clean_one_time_passwords')
        assert_equal(OneTimePassword.objects.filter(pk__in=[obj.pk for obj in old_expired_otp]).count(), 0)
        assert_equal(OneTimePassword.objects.filter(pk__in=[obj.pk for obj in new_expired_otp]).count(), 10)
        assert_equal(OneTimePassword.objects.filter(pk__in=[obj.pk for obj in inactive_otp]).count(), 0)
        assert_equal(OneTimePassword.objects.filter(pk__in=[obj.pk for obj in not_expired_otp]).count(), 10)




# Generated by Django 3.1 on 2021-01-19 12:08

from datetime import timedelta

from django.conf import settings
from django.db import migrations, models
from django.db.models import F


def set_mobile_device_changed_at(apps, schema_editor):
    MobileDevice = apps.get_model('auth_token', 'MobileDevice')
    MobileDevice.objects.filter(last_login__isnull=True).update(changed_at=F('created_at'))
    MobileDevice.objects.filter(last_login__isnull=False).update(changed_at=F('last_login'))


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
        ('auth_token', '0011_auto_20210119_1308'),
    ]

    operations = [
        migrations.RunPython(set_mobile_device_changed_at),
    ]



import string

from datetime import timedelta

from unittest import mock

from freezegun import freeze_time

from django.core.management import call_command
from django.utils.timezone import now
from django.test import override_settings
from django.test.client import RequestFactory

from germanium.decorators import data_consumer
from germanium.test_cases.default import GermaniumTestCase
from germanium.tools import (
    assert_equal, assert_true, assert_raises, assert_false, assert_not_raises, assert_not_equal, assert_is_none,
    assert_is_instance, assert_is_not_none
)

from auth_token.enums import AuthorizationRequestState
from auth_token.exceptions import KeyGeneratorError
from auth_token.models import AnonymousUser, AnonymousAuthorizationToken, MobileDevice
from auth_token.utils import (
    compute_expires_at, hash_key, header_name_to_django, generate_key,
    create_auth_header_value, login, logout, parse_auth_header_value, get_token_key_from_request,
    get_token, dont_enforce_csrf_checks, get_user_from_token, get_user, takeover, create_otp, deactivate_otp,
    get_valid_otp, check_otp, extend_otp, create_authorization_request, check_authorization_request,
    grant_authorization_request, deny_authorization_request,cancel_authorization_request, reset_authorization_request,
    get_otp_qs
)
from auth_token.signals import authorization_granted, authorization_denied, authorization_cancelled

from .base import BaseTestCaseMixin


__all__ = (
   'UtilsTestCase',
)


def authorization_receiver(sender, authorization_request, **kwargs):
    pass


def test_otp_generator():
    return '12345'


class UtilsTestCase(BaseTestCaseMixin, GermaniumTestCase):

    def set_up(self):
        # Every test needs access to the request factory.
        self.requiest_factory = RequestFactory()

    @freeze_time(now())
    def test_compute_expires_at_should_return_valid_value(self):
        assert_equal(compute_expires_at(61), now() + timedelta(seconds=61))
        assert_equal(compute_expires_at(126), now() + timedelta(seconds=126))

    test_hash_data = (
        ('1234', '', '2accd14793919bb56639d6e90992378bd5dcb1384082a0822f851aff73587c14'),
        ('secret', '', '829dd60c4cb2c9acd3068bb6f42af56047a0c677067dcf55e26bc7755d7f193c'),
        ('secret', None, '829dd60c4cb2c9acd3068bb6f42af56047a0c677067dcf55e26bc7755d7f193c'),
        ('secret', 'salt', '8ed3095217b32faeafbb2b85463c7de48827df6219a4eaaf4268e4220f6085a9'),
    )

    @data_consumer(test_hash_data)
    def test_hash_key_should_return_valid_value(self, key, salt, key_hash):
        assert_equal(hash_key(key, salt=salt), key_hash)

    test_header_names_data = (
        ('X-Auth', 'HTTP_X_AUTH'),
        ('X-AUTH', 'HTTP_X_AUTH'),
        ('x-auth', 'HTTP_X_AUTH'),
    )

    @data_consumer(test_header_names_data)
    def test_header_name_to_django_should_return_valid_value(self, header_name, django_header_name):
        assert_equal(header_name_to_django(header_name), django_header_name)

    def test_generate_key_should_generate_random_keys(self):
        assert_equal(generate_key(characters='a'), 20 * 'a')
        assert_equal(generate_key(characters='a', length=10), 10 * 'a')
        assert_equal(len(generate_key(length=10)), 10)
        assert_true(set(generate_key(characters='ab')) <= {'a', 'b'})

    def test_create_auth_header_value_should_return_value_with_bearer_prefix(self):
        assert_equal(create_auth_header_value('token'), 'Bearer token')

    @data_consumer('create_user')
    def test_login_user_should_require_select_backend(self, user):
        request = self.requiest_factory.get('/login/')
        with assert_raises(ValueError):
            login(request, user)

    @freeze_time(now())
    @data_consumer('create_user')
    def test_login_user_should_create_authorization_token(self, user):
        request = self.requiest_factory.get('/login/')
        login(request, user, backend='LoginBackend')
        authorization_token = request.token
        assert_equal(authorization_token.user, user)
        assert_true(authorization_token.is_authenticated)
        assert_equal(authorization_token.ip, '127.0.0.1')
        assert_equal(authorization_token.expires_at, now() + timedelta(hours=1))
        assert_equal(authorization_token.backend, 'LoginBackend')
        assert_true(authorization_token.allowed_cookie)
        assert_true(authorization_token.allowed_header)

    @freeze_time(now())
    @data_consumer('create_user')
    def test_login_user_with_two_factor_should_create_unauthenticated_authorization(self, user):
        request = self.requiest_factory.get('/login/')
        login(request, user, backend='LoginBackend', two_factor_login=True)
        assert_false(request.token.is_authenticated)

    @freeze_time(now())
    @data_consumer('create_user')
    def test_login_user_with_no_default_values_should_create_authorization(self, user):
        request = self.requiest_factory.get('/login/')
        login(request, user, backend='LoginBackend', two_factor_login=True, allowed_cookie=False,
              allowed_header=False, expiration=1, auth_slug='test', related_objs=[user])
        authorization_token = request.token
        assert_false(authorization_token.is_authenticated)
        assert_equal(authorization_token.ip, '127.0.0.1')
        assert_equal(authorization_token.expires_at, now() + timedelta(seconds=1))
        assert_equal(authorization_token.backend, 'LoginBackend')
        assert_false(authorization_token.allowed_cookie)
        assert_false(authorization_token.allowed_header)
        assert_true(authorization_token.related_objects.count(), 1)
        assert_equal(authorization_token.related_objects.get().object, user)

    def test_logout_without_user_should_not_raise_exception(self):
        with assert_not_raises(Exception):
            request = self.requiest_factory.get('/logout/')
            logout(request)

    @data_consumer('create_user')
    def test_logout_with_user_without_token_should_logout_it(self, user):
        request = self.requiest_factory.get('/logout/')
        request.user = user
        logout(request)
        assert_false(request.user.is_authenticated)
        assert_not_equal(request.user, user)

    @data_consumer('create_user')
    def test_logout_with_user_with_token_should_logout_it(self, user):
        request = self.requiest_factory.get('/logout/')
        request.user = None
        login(request, user, backend='LoginBackend')
        assert_true(request.user.is_authenticated)
        assert_true(request.token.is_authenticated)
        logout(request)
        assert_false(request.user.is_authenticated)
        assert_false(request.token.is_authenticated)

    def test_parse_auth_header_value_should_return_token(self):
        request = self.requiest_factory.get('/')
        request.META['HTTP_AUTHORIZATION'] = 'Bearer TOKEN'
        assert_equal(parse_auth_header_value(request), 'TOKEN')

    def test_parse_auth_header_missing_value_should_raise_exception(self):
        request = self.requiest_factory.get('/')
        with assert_raises(ValueError):
            parse_auth_header_value(request)

    def test_parse_auth_header_invalid_value_should_return_none(self):
        request = self.requiest_factory.get('/')
        request.META['HTTP_AUTHORIZATION'] = 'Invalid TOKEN'
        assert_is_none(parse_auth_header_value(request))

    def test_get_token_key_from_request_should_be_returned_from_header(self):
        request = self.requiest_factory.get('/')
        request.META['HTTP_AUTHORIZATION'] = 'Bearer TOKEN'
        assert_equal(get_token_key_from_request(request), ('TOKEN', True, False))

    def test_get_token_key_from_request_should_be_returned_from_cookie(self):
        request = self.requiest_factory.get('/')
        request.COOKIES['Authorization'] = 'TOKEN'
        assert_equal(get_token_key_from_request(request), ('TOKEN', False, True))

    def test_get_token_key_from_request_should_return_no_token(self):
        request = self.requiest_factory.get('/')
        assert_equal(get_token_key_from_request(request), (None, False, False))

    @data_consumer('create_user')
    def test_get_token_should_return_token_from_request_header(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='LoginBackend')
        token = request.token
        request.META['HTTP_AUTHORIZATION'] = 'Bearer {}'.format(token.secret_key)
        assert_equal(get_token(request), token)

    @data_consumer('create_user')
    def test_get_token_should_return_token_from_request_cookie(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='LoginBackend')
        token = request.token
        request.COOKIES['Authorization'] = token.secret_key
        assert_equal(get_token(request), token)

    @data_consumer('create_user')
    def test_get_token_should_not_return_token_from_header_if_secret_is_invalid(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='LoginBackend')
        request.META['HTTP_AUTHORIZATION'] = 'Bearer invalid'
        assert_false(get_token(request).is_active)
        assert_not_equal(get_token(request), request.token)

    @data_consumer('create_user')
    def test_get_token_should_not_return_token_from_cookie_if_secret_is_invalid(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='LoginBackend')
        request.COOKIES['Authorization'] = 'invalid'
        assert_false(get_token(request).is_active)
        assert_not_equal(get_token(request), request.token)

    @data_consumer('create_user')
    def test_get_token_should_not_return_token_without_authorization_data_in_request(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='LoginBackend')
        assert_false(get_token(request).is_active)
        assert_not_equal(get_token(request), request.token)

    @data_consumer('create_user')
    def test_get_token_should_not_return_token_from_request_cookie_if_cookie_is_not_allowed(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='LoginBackend', allowed_cookie=False)
        token = request.token
        request.COOKIES['Authorization'] = token.secret_key
        assert_false(get_token(request).is_active)
        assert_not_equal(get_token(request), request.token)

    @data_consumer('create_user')
    def test_get_token_should_not_return_token_from_request_header_if_header_is_not_allowed(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='LoginBackend', allowed_header=False)
        token = request.token
        request.META['HTTP_AUTHORIZATION'] = 'Bearer {}'.format(token.secret_key)
        assert_false(get_token(request).is_active)
        assert_not_equal(get_token(request), request.token)

    @data_consumer('create_user')
    def test_get_token_should_not_return_token_from_request_header_if_token_is_expired(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='LoginBackend')
        token = request.token
        token.change_and_save(expires_at=now())
        request.META['HTTP_AUTHORIZATION'] = 'Bearer {}'.format(token.secret_key)
        assert_false(get_token(request).is_active)
        assert_not_equal(get_token(request), request.token)

    def test_dont_enforce_csrf_checks_should_return_false_for_no_token(self):
        request = self.requiest_factory.get('/')
        assert_false(dont_enforce_csrf_checks(request))

    @data_consumer('create_user')
    def test_dont_enforce_csrf_checks_should_return_true_for_token_from_header(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='LoginBackend')
        token = request.token
        request.META['HTTP_AUTHORIZATION'] = 'Bearer {}'.format(token.secret_key)
        assert_true(dont_enforce_csrf_checks(request))

    def test_dont_enforce_csrf_checks_should_return_true_for_request_with_flag(self):
        request = self.requiest_factory.get('/')
        request._dont_enforce_csrf_checks = True
        assert_true(dont_enforce_csrf_checks(request))

    def test_get_user_from_token_should_return_anonymous_user_for_no_token_or_anonymous_token(self):
        assert_is_instance(get_user_from_token(None), AnonymousUser)
        assert_is_instance(get_user_from_token(AnonymousAuthorizationToken()), AnonymousUser)

    @data_consumer('create_user')
    def test_get_user_from_token_should_return_anonymous_user_for_token_with_invalid_backend(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='LoginBackend')
        token = request.token
        request.META['HTTP_AUTHORIZATION'] = 'Bearer {}'.format(token.secret_key)
        assert_is_instance(get_user_from_token(token), AnonymousUser)

    @data_consumer('create_user')
    def test_get_user_from_token_should_return_user_for_valid_token(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='auth_token.backends.DeviceBackend')
        token = request.token
        request.META['HTTP_AUTHORIZATION'] = 'Bearer {}'.format(token.secret_key)
        assert_equal(get_user_from_token(token), user)

    @data_consumer('create_user')
    def test_get_user_should_return_user_for_valid_token(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='auth_token.backends.DeviceBackend')
        request.META['HTTP_AUTHORIZATION'] = 'Bearer {}'.format(request.token.secret_key)
        assert_equal(get_user(request), user)

    @data_consumer('create_user')
    def test_takeover_same_user_should_not_create_takeover_instance_and_return_false(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='auth_token.backends.DeviceBackend')
        request.user = user
        request.META['HTTP_AUTHORIZATION'] = 'Bearer {}'.format(request.token.secret_key)
        assert_false(takeover(request, user))
        assert_equal(request.token.user_takeovers.count(), 0)

    @data_consumer('create_user')
    def test_takeover_user_should_create_takeover_instance_and_return_true(self, user):
        request = self.requiest_factory.get('/')
        login(request, user, backend='auth_token.backends.DeviceBackend')
        request.user = user
        assert_equal(get_user_from_token(request.token), user)
        request.META['HTTP_AUTHORIZATION'] = 'Bearer {}'.format(request.token.secret_key)
        takeovered_user = self.create_user(username='takeovered', email='takeover@test.cz')
        assert_true(takeover(request, takeovered_user))
        assert_equal(request.token.user_takeovers.filter(is_active=True).count(), 1)
        assert_equal(get_user_from_token(request.token), takeovered_user)
        logout(request)
        assert_equal(request.token.user_takeovers.filter(is_active=True).count(), 0)
        assert_equal(get_user_from_token(request.token), user)
        logout(request)
        assert_is_instance(get_user_from_token(request.token), AnonymousUser)

    @freeze_time(now())
    def test_create_otp_should_create_new_otp_instance_with_default_expiration_time(self):
        otp = create_otp('test', key_generator=lambda: '1234')
        assert_equal(otp.slug, 'test')
        assert_equal(otp.expires_at, now() + timedelta(hours=1))
        assert_equal(otp.secret_key, '1234')
        assert_equal(otp.key, hash_key('1234', salt='test'))

    @freeze_time(now())
    def test_create_otp_should_create_new_otp_instance_with_default_expiration_time(self):
        otp = create_otp('test')
        assert_equal(otp.slug, 'test')
        assert_equal(otp.expires_at, now() + timedelta(hours=1))
        assert_equal(otp.key, hash_key(otp.secret_key, salt='test'))

    @freeze_time(now())
    @data_consumer('create_user')
    def test_create_otp_should_create_new_otp_with_custom_key_generator_and_expiration(self, user):
        otp = create_otp('test', key_generator=lambda: '1234', expiration=14, related_objects=[user],
                         data={'otp': 'data'})
        assert_equal(otp.slug, 'test')
        assert_equal(otp.expires_at, now() + timedelta(seconds=14))
        assert_equal(otp.secret_key, '1234')
        assert_equal(otp.key, hash_key('1234', salt='test'))
        assert_equal(otp.data, {'otp': 'data'})
        assert_equal(otp.related_objects.count(), 1)
        assert_equal(otp.related_objects.get().object, user)

    @freeze_time(now())
    @override_settings(AUTH_TOKEN_OTP_DEFAULT_KEY_GENERATOR='app.tests.utils.test_otp_generator')
    @data_consumer('create_user')
    def test_create_otp_should_create_new_otp_with_changed_default_generator_in_config(self, user):
        otp = create_otp('test')
        assert_equal(otp.secret_key, '12345')
        assert_equal(otp.key, hash_key('12345', salt='test'))

    def test_create_otp_with_deactivation_old_otp_codes_should_deactivate_it(self):
        otp1 = create_otp('test')
        otp2 = create_otp('test')
        otp3 = create_otp('test2')

        assert_true(otp1.refresh_from_db().is_active)
        assert_true(otp2.refresh_from_db().is_active)
        assert_true(otp3.refresh_from_db().is_active)

        otp4 = create_otp('test', deactivate_old=True)

        assert_false(otp1.refresh_from_db().is_active)
        assert_false(otp2.refresh_from_db().is_active)
        assert_true(otp3.refresh_from_db().is_active)
        assert_true(otp4.refresh_from_db().is_active)

    @data_consumer('create_user')
    def test_deactivate_otp_should_deactivate_otp_according_to_input(self, user):
        otp1 = create_otp('test')
        otp2 = create_otp('test')
        otp3 = create_otp('test', related_objects=[user])
        otp4 = create_otp('test2', related_objects=[user])
        otp5 = create_otp('test')

        deactivate_otp('test', key=otp1.secret_key)
        assert_false(otp1.refresh_from_db().is_active)
        assert_true(otp2.refresh_from_db().is_active)
        assert_true(otp3.refresh_from_db().is_active)
        assert_true(otp4.refresh_from_db().is_active)
        assert_true(otp5.refresh_from_db().is_active)

        deactivate_otp('test', related_objects=[user])
        assert_false(otp1.refresh_from_db().is_active)
        assert_true(otp2.refresh_from_db().is_active)
        assert_false(otp3.refresh_from_db().is_active)
        assert_true(otp4.refresh_from_db().is_active)
        assert_true(otp5.refresh_from_db().is_active)

        deactivate_otp('test')
        assert_false(otp1.refresh_from_db().is_active)
        assert_false(otp2.refresh_from_db().is_active)
        assert_false(otp3.refresh_from_db().is_active)
        assert_true(otp4.refresh_from_db().is_active)
        assert_false(otp5.refresh_from_db().is_active)

    def test_get_valid_otp_should_return_only_valid_otp(self):
        otp1 = create_otp('test')
        otp2 = create_otp('test2')

        assert_is_none(get_valid_otp('test', 'invalid'))
        assert_is_none(get_valid_otp('test2', otp1.secret_key))

        assert_equal(get_valid_otp('test', otp1.secret_key), otp1)
        assert_equal(get_valid_otp('test2', otp2.secret_key), otp2)

        deactivate_otp('test')
        assert_is_none(get_valid_otp('test', otp1.secret_key))

    def test_get_valid_otp_should_return_last_valid_otp(self):
        otp1 = create_otp('test')
        otp2 = create_otp('test')

        assert_equal(get_valid_otp('test', otp1.secret_key), otp1)
        assert_equal(get_valid_otp('test', otp2.secret_key), otp2)
        assert_equal(get_valid_otp('test'), otp2)

    @data_consumer('create_user')
    def test_get_valid_otp_should_return_otp_according_to_related_objects(self, user):
        otp1 = create_otp('test', related_objects=[user])
        otp2 = create_otp('test')

        assert_equal(get_valid_otp('test', otp1.secret_key), otp1)
        assert_equal(get_valid_otp('test', otp2.secret_key), otp2)
        assert_equal(get_valid_otp('test'), otp2)
        assert_equal(get_valid_otp('test', related_objects=[user]), otp1)
        assert_is_none(get_valid_otp('test', otp2.secret_key, related_objects=[user]))

    def test_get_valid_otp_should_deactive_returned_otp(self):
        otp = create_otp('test')
        get_valid_otp('test', otp.secret_key, deactivate=True)
        assert_false(otp.refresh_from_db().is_active)

    def test_check_otp_should_return_true_if_input_data_are_valid(self):
        otp1 = create_otp('test')
        otp2 = create_otp('test2')

        assert_false(check_otp('test', 'invalid'))
        assert_false(check_otp('test2', otp1.secret_key))

        assert_true(check_otp('test', otp1.secret_key))
        assert_true(check_otp('test2', otp2.secret_key))

        deactivate_otp('test')
        assert_false(check_otp('test', otp1.secret_key))

    @freeze_time(now())
    def test_extend_otp_should_extend_otp_expiration(self):
        otp = create_otp('test', expiration=1)
        assert_equal(otp.expires_at, now() + timedelta(seconds=1))

        assert_true(extend_otp('test', otp.secret_key, expiration=5))
        assert_equal(otp.refresh_from_db().expires_at, now() + timedelta(seconds=5))

        assert_true(extend_otp('test', otp.secret_key, expiration=3600))
        assert_equal(otp.refresh_from_db().expires_at, now() + timedelta(hours=1))

        assert_false(extend_otp('test2', otp.secret_key, expiration=3611))
        assert_equal(otp.refresh_from_db().expires_at, now() + timedelta(hours=1))

        assert_false(extend_otp('test', 'invalid', expiration=3611))
        assert_equal(otp.refresh_from_db().expires_at, now() + timedelta(hours=1))

        deactivate_otp('test')
        assert_false(extend_otp('test', otp.secret_key, expiration=3611))
        assert_equal(otp.refresh_from_db().expires_at, now() + timedelta(hours=1))

    @freeze_time(now())
    @data_consumer('create_user')
    @override_settings(AUTH_TOKEN_AUTHORIZATION_REQUEST_BACKENDS=[
        'auth_token.authorization_request.backends.OTPAuthorizationRequestBackend'])
    def test_create_authorization_request_should_create_new_authorization_with_otp(self, user):
        authorization_request = create_authorization_request('test', user, 'test')
        assert_equal(authorization_request.slug, 'test')
        assert_equal(authorization_request.title, 'test')
        assert_is_none(authorization_request.description)
        assert_equal(get_otp_qs(related_objects=[authorization_request], is_active=True).count(), 1)
        assert_equal(authorization_request.expires_at, now() + timedelta(hours=1))

    @freeze_time(now())
    @data_consumer('create_user')
    def test_create_authorization_request_should_create_new_authorization_with_no_default_values(self, user):
        authorization_request = create_authorization_request(
            'slug', user, 'title', description='description', related_objects=[user],
            data={'authorization': 'data'},  expiration=1,
            backend_path='auth_token.authorization_request.backends.OTPAuthorizationRequestBackend'
        )
        assert_equal(authorization_request.slug, 'slug')
        assert_equal(authorization_request.user, user)
        assert_equal(authorization_request.title, 'title')
        assert_equal(authorization_request.description, 'description')
        assert_equal(authorization_request.related_objects.count(), 1)
        assert_equal(authorization_request.related_objects.get().object, user)
        assert_equal(
            get_valid_otp(authorization_request.slug, related_objects=[authorization_request]).expires_at,
            now() + timedelta(seconds=1)
        )
        assert_equal(authorization_request.expires_at, now() + timedelta(seconds=1))

    @data_consumer('create_user')
    def test_create_authorization_request_with_device_type_should_create_it(self, user):
        authorization_request = create_authorization_request(
            'slug', user, 'title',
            backend_path='auth_token.authorization_request.backends.MobileDeviceAuthorizationRequestBackend'
        )
        assert_equal(authorization_request.slug, 'slug')
        assert_equal(authorization_request.title, 'title')

    @data_consumer('create_user')
    def test_check_authorization_request_with_mobile_device_type_should_return_right_value(self, user):
        mobile_device_login_token = MobileDevice.objects.activate_or_create(
            uuid='E621E1F8C36C495', user=user
        ).secret_password
        authorization_request = create_authorization_request(
            'slug', user, 'title',
            backend_path='auth_token.authorization_request.backends.MobileDeviceAuthorizationRequestBackend'
        )

        assert_false(check_authorization_request(authorization_request))
        assert_false(check_authorization_request(authorization_request, mobile_device_uuid='E621E1F8C36C495'))
        assert_true(check_authorization_request(authorization_request, mobile_device_uuid='E621E1F8C36C495',
                                                mobile_login_token=mobile_device_login_token))

    @data_consumer('create_user')
    def test_check_authorization_request_with_mobile_device_type_and_token_should_return_right_value(self, user):
        mobile_device = MobileDevice.objects.activate_or_create(
            uuid='E621E1F8C36C495', user=user
        )
        authorization_request = create_authorization_request(
            'slug', user, 'title',
            backend_path='auth_token.authorization_request.backends.MobileDeviceAuthorizationRequestBackend'
        )
        assert_true(check_authorization_request(authorization_request, mobile_device_uuid='E621E1F8C36C495',
                                                mobile_login_token=mobile_device.secret_password))

    @freeze_time(now())
    @data_consumer('create_user')
    def test_check_authorization_request_with_otp_should_return_right_value(self, user):
        authorization_request = create_authorization_request(
            'test', user, 'test',
            backend_path='auth_token.authorization_request.backends.OTPAuthorizationRequestBackend'
        )

        assert_equal(authorization_request.state, AuthorizationRequestState.WAITING)
        assert_false(check_authorization_request(authorization_request))
        assert_false(check_authorization_request(authorization_request, otp_secret_key='invalid'))
        assert_true(check_authorization_request(authorization_request,
                                                otp_secret_key=authorization_request.secret_key))
        with override_settings(AUTH_TOKEN_AUTHORIZATION_REQUEST_OTP_DEBUG_CODE='invalid'):
            assert_true(check_authorization_request(authorization_request, otp_secret_key='invalid'))

        with freeze_time(now() + timedelta(hours=1, seconds=1)):
            # Expired authorization
            assert_false(check_authorization_request(authorization_request,
                                                     otp_secret_key=authorization_request.secret_key))
            assert_equal(authorization_request.state, AuthorizationRequestState.EXPIRED)

    @freeze_time(now())
    @data_consumer('create_user')
    def test_authorization_reset_create_new_otp_key_and_increase_expiration(self, user):
        authorization_request = create_authorization_request(
            'test', user, 'test',
            backend_path='auth_token.authorization_request.backends.OTPAuthorizationRequestBackend'
        )
        assert_equal(get_otp_qs(related_objects=[authorization_request]).count(), 1)
        assert_equal(authorization_request.expires_at, now() + timedelta(hours=1))

        secret_key = authorization_request.secret_key
        assert_true(check_authorization_request(authorization_request, otp_secret_key=secret_key))

        with freeze_time(now() + timedelta(minutes=5)):
            authorization_request = reset_authorization_request(authorization_request)
            assert_equal(authorization_request.expires_at, now() + timedelta(hours=1))
            assert_equal(get_otp_qs(related_objects=[authorization_request]).count(), 2)
            assert_equal(get_otp_qs(related_objects=[authorization_request], is_active=True).count(), 1)
            assert_false(check_authorization_request(authorization_request, otp_secret_key=secret_key))
            assert_true(check_authorization_request(authorization_request,
                                                    otp_secret_key=authorization_request.secret_key))

            authorization_request = reset_authorization_request(authorization_request, expiration=5)
            assert_equal(authorization_request.expires_at, now() + timedelta(seconds=5))

    @data_consumer('create_user')
    def test_grant_authorization_request_should_grant_authorization_and_call_receiver(self, user):
        authorization_request = create_authorization_request(
            'test', user, 'test',
            backend_path='auth_token.authorization_request.backends.OTPAuthorizationRequestBackend'
        )
        assert_equal(authorization_request.state, AuthorizationRequestState.WAITING)
        with mock.patch('dj.apps.app.tests.utils.authorization_receiver') as mocked_receiver:
            authorization_granted.connect(mocked_receiver, sender='test')
            grant_authorization_request(authorization_request, extra_data='test')
            mocked_receiver.assert_called()
            assert_equal(authorization_request.refresh_from_db().state, AuthorizationRequestState.GRANTED)

    @data_consumer('create_user')
    def test_grant_authorization_request_should_grant_authorization_and_not_call_receiver_with_another_slug(self, user):
        authorization_request = create_authorization_request(
            'test', user, 'test',
            backend_path='auth_token.authorization_request.backends.OTPAuthorizationRequestBackend'
        )
        assert_equal(authorization_request.state, AuthorizationRequestState.WAITING)
        with mock.patch('dj.apps.app.tests.utils.authorization_receiver') as mocked_receiver:
            authorization_granted.connect(mocked_receiver, sender='test2')
            grant_authorization_request(authorization_request)
            mocked_receiver.assert_not_called()
            authorization_request.refresh_from_db()
            assert_equal(authorization_request.state, AuthorizationRequestState.GRANTED)
            assert_is_not_none(authorization_request.granted_at)

    @data_consumer('create_user')
    def test_grant_authorization_request_should_not_grant_already_granted_authorization(self, user):
        authorization_request = create_authorization_request(
            'test', user, 'test',
            backend_path='auth_token.authorization_request.backends.OTPAuthorizationRequestBackend'
        )
        assert_equal(authorization_request.state, AuthorizationRequestState.WAITING)
        grant_authorization_request(authorization_request)
        assert_equal(authorization_request.refresh_from_db().state, AuthorizationRequestState.GRANTED)
        with assert_raises(AssertionError):
            grant_authorization_request(authorization_request)

    @data_consumer('create_user')
    def test_deny_authorization_request_should_not_deny_already_granted_authorization(self, user):
        authorization_request = create_authorization_request(
            'test', user, 'test',
            backend_path='auth_token.authorization_request.backends.OTPAuthorizationRequestBackend'
        )
        assert_equal(authorization_request.state, AuthorizationRequestState.WAITING)
        grant_authorization_request(authorization_request)
        assert_equal(authorization_request.refresh_from_db().state, AuthorizationRequestState.GRANTED)
        with assert_raises(AssertionError):
            deny_authorization_request(authorization_request)

    @data_consumer('create_user')
    def test_deny_authorization_request_should_deny_authorization_and_call_receiver(self, user):
        authorization_request = create_authorization_request(
            'test', user, 'test',
            backend_path='auth_token.authorization_request.backends.OTPAuthorizationRequestBackend'
        )
        assert_equal(authorization_request.state, AuthorizationRequestState.WAITING)
        with mock.patch('dj.apps.app.tests.utils.authorization_receiver') as mocked_receiver:
            authorization_denied.connect(mocked_receiver, sender='test')
            deny_authorization_request(authorization_request)
            mocked_receiver.assert_called()
            assert_equal(authorization_request.refresh_from_db().state, AuthorizationRequestState.DENIED)

    @data_consumer('create_user')
    def test_deny_authorization_request_should_deny_authorization_and_not_call_receiver_with_another_slug(self, user):
        authorization_request = create_authorization_request(
            'test', user, 'test',
            backend_path='auth_token.authorization_request.backends.OTPAuthorizationRequestBackend'
        )
        assert_equal(authorization_request.state, AuthorizationRequestState.WAITING)
        with mock.patch('dj.apps.app.tests.utils.authorization_receiver') as mocked_receiver:
            authorization_denied.connect(mocked_receiver, sender='test2')
            deny_authorization_request(authorization_request)
            mocked_receiver.assert_not_called()
            assert_equal(authorization_request.refresh_from_db().state, AuthorizationRequestState.DENIED)

    @data_consumer('create_user')
    def test_cancel_authorization_request_should_not_cancel_already_granted_authorization(self, user):
        authorization_request = create_authorization_request(
            'test', user, 'test',
            backend_path='auth_token.authorization_request.backends.OTPAuthorizationRequestBackend'
        )
        assert_equal(authorization_request.state, AuthorizationRequestState.WAITING)
        grant_authorization_request(authorization_request)
        assert_equal(authorization_request.refresh_from_db().state, AuthorizationRequestState.GRANTED)
        with assert_raises(AssertionError):
            cancel_authorization_request(authorization_request)

    @data_consumer('create_user')
    def test_cancel_authorization_request_should_deny_authorization_and_call_receiver(self, user):
        authorization_request = create_authorization_request(
            'test', user, 'test',
            backend_path='auth_token.authorization_request.backends.OTPAuthorizationRequestBackend'
        )
        assert_equal(authorization_request.state, AuthorizationRequestState.WAITING)
        with mock.patch('dj.apps.app.tests.utils.authorization_receiver') as mocked_receiver:
            authorization_cancelled.connect(mocked_receiver, sender='test')
            cancel_authorization_request(authorization_request)
            mocked_receiver.assert_called()
            assert_equal(authorization_request.refresh_from_db().state, AuthorizationRequestState.CANCELLED)

    @data_consumer('create_user')
    def test_cancel_authorization_request_should_cancel_authorization_and_not_call_receiver_with_another_slug(self,
                                                                                                              user):
        authorization_request = create_authorization_request(
            'test', user, 'test',
            backend_path='auth_token.authorization_request.backends.OTPAuthorizationRequestBackend'
        )
        assert_equal(authorization_request.state, AuthorizationRequestState.WAITING)
        with mock.patch('dj.apps.app.tests.utils.authorization_receiver') as mocked_receiver:
            authorization_cancelled.connect(mocked_receiver, sender='test2')
            cancel_authorization_request(authorization_request)
            mocked_receiver.assert_not_called()
            assert_equal(authorization_request.refresh_from_db().state, AuthorizationRequestState.CANCELLED)

    @freeze_time(now())
    def test_duplicite_key_generator_should_raise_exception(self):
        create_otp('test', key_generator=lambda: '1234')
        with assert_raises(KeyGeneratorError):
            create_otp('test', key_generator=lambda: '1234')

    @freeze_time(now())
    def test_key_genrator_should_be_called_more_times(self):
        def key_genertor():
            key_genertor.iteration += 1
            if key_genertor.iteration < 100:
                return '1234'
            else:
                return '1235'
        key_genertor.iteration = 0

        create_otp('test', key_generator=lambda: '1234')
        assert_equal(create_otp('test', key_generator=key_genertor).secret_key, '1235')
        assert_equal(key_genertor.iteration, 100)



from datetime import timedelta

from django.conf import settings as django_settings
from django.contrib.auth.hashers import check_password, make_password
from django.contrib.auth.models import AnonymousUser
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models, IntegrityError, transaction
from django.db.utils import IntegrityError
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from enumfields import NumEnumField

from chamber.exceptions import PersistenceException
from chamber.models import SmartModel, SmartQuerySet, SmartManager

from generic_m2m_field.models import GenericManyToManyField

from auth_token.utils import compute_expires_at, generate_key, hash_key
from auth_token.config import settings

from .enums import AuthorizationRequestState, AuthorizationRequestResult
from .exceptions import KeyGeneratorError


KEY_SALT = 'django-auth-token'


def compute_authorization_token_expires_at(expiration=None):
    return compute_expires_at(expiration or settings.DEFAULT_TOKEN_AGE)


def generate_token_key():
    return generate_key(length=settings.LENGTH)


class BaseHashKeyManager(SmartManager):

    def _hash_key(self, key, **kwargs):
        return hash_key(key)

    def create(self, key_generator, **kwargs):
        for _ in range(settings.MAX_RANDOM_KEY_ITERATIONS):
            try:
                key = key_generator()
                hashed_key = self._hash_key(key, **kwargs)
                obj = super().create(key=hashed_key, **kwargs)
                obj.secret_key = key
                return obj
            except IntegrityError:
                pass
            except PersistenceException as ex:
                if not ex.error_dict and 'key' not in ex.error_dict:
                    raise ex
        raise KeyGeneratorError('Could not produce unique key')


class AuthorizationTokenManager(BaseHashKeyManager):

    def create(self, **kwargs):
        return super().create(generate_token_key, **kwargs)


class AuthorizationToken(SmartModel):
    """
    The default authorization token model.
    """

    key = models.CharField(
        verbose_name=_('key'),
        max_length=128,
        unique=True,
        db_index=True,
        null=False,
        blank=False
    )
    user = models.ForeignKey(
        verbose_name=_('user'),
        to=django_settings.AUTH_USER_MODEL,
        related_name='authorization_tokens',
        null=False,
        blank=False,
        on_delete=models.CASCADE
    )
    is_active = models.BooleanField(
        verbose_name=_('is active'),
        default=True
    )
    user_agent = models.CharField(
        verbose_name=_('user agent'),
        max_length=256,
        null=True,
        blank=True
    )
    expires_at = models.DateTimeField(
        verbose_name=_('expires at'),
        null=False,
        blank=False,
        default=compute_authorization_token_expires_at
    )
    ip = models.GenericIPAddressField(
        verbose_name=_('IP'),
        null=False,
        blank=False
    )
    auth_slug = models.SlugField(
        verbose_name=_('slug'),
        null=True,
        blank=True
    )
    backend = models.CharField(
        verbose_name=_('backend'),
        max_length=250,
        null=False,
        blank=False
    )
    allowed_cookie = models.BooleanField(
        verbose_name=_('is allowed cookie'),
        default=True
    )
    allowed_header = models.BooleanField(
        verbose_name=_('is allowed header'),
        default=True
    )
    is_authenticated = models.BooleanField(
        verbose_name=_('is authenticated'),
        null=False,
        blank=False,
        default=False
    )
    preserve_cookie = models.BooleanField(
        verbose_name=_('preserve cookie'),
        null=False,
        blank=False,
        default=False
    )
    mobile_device = models.ForeignKey(
        verbose_name=_('mobile device'),
        to='MobileDevice',
        related_name='authorization_tokens',
        null=True,
        blank=True,
        on_delete=models.CASCADE
    )

    related_objects = GenericManyToManyField()

    is_from_header = False
    is_from_cookie = False
    secret_key = None
    is_anonymous = False

    objects = AuthorizationTokenManager()

    class Meta:
        verbose_name = _('authorization token')
        verbose_name_plural = _('authorization tokens')

    @property
    def active_takeover(self):
        return self.user_takeovers.filter(is_active=True).last()

    @property
    def is_expired(self):
        return self.expires_at < timezone.now()

    @property
    def time_to_expiration(self):
        return max(timedelta(seconds=0), self.expires_at - timezone.now())


class AnonymousAuthorizationToken:

    key = None
    user = AnonymousUser
    created_at = None
    is_active = False
    user_agent = None
    expiration = None
    is_from_header = False
    is_from_cookie = False
    active_takeover = None
    backend = None
    allowed_cookie = False
    allowed_header = False
    secret_key = None
    is_authenticated = False
    is_anonymous = True

    def save(self):
        raise NotImplementedError

    def delete(self):
        raise NotImplementedError


class UserAuthorizationTokenTakeover(SmartModel):
    """
    The model allows to change user without token change
    """

    token = models.ForeignKey(
        verbose_name=_('authorization token'),
        to=AuthorizationToken,
        related_name='user_takeovers',
        on_delete=models.CASCADE
    )
    user = models.ForeignKey(
        verbose_name=_('user'),
        to=django_settings.AUTH_USER_MODEL,
        related_name='user_token_takeovers',
        null=False,
        blank=False,
        on_delete=models.CASCADE
    )
    is_active = models.BooleanField()

    class Meta:
        verbose_name = _('authorization takeover')
        verbose_name_plural = _('authorization takeovers')


class MobileDeviceAlreadyExists(Exception):
    pass


class MobileDeviceQuerySet(SmartQuerySet):

    def activate_or_create(self, uuid, user, user_agent='', name=None, slug=None, is_primary=True):
        """
        This method must be called when user is authenticated.
        It creates a new MobileDevice with auto generated token for the device and returns token.
        If MobileDevice with same UUID exists MobileDeviceAlreadyExists is raised.
        """
        secret_password = generate_key(length=settings.MOBILE_DEVICE_SECRET_PASSWORD_LENGTH)
        defaults = {
            'login_token': make_password(secret_password),
            'user_agent': user_agent[:256],
            'is_active': True,
            'name': name,
            'slug': slug,
            'is_primary': is_primary
        }
        mobile_device, is_created_mobile_device = self.get_or_create(
            uuid=uuid,
            user=user,
            defaults=defaults
        )
        mobile_device.secret_password = secret_password
        if not is_created_mobile_device and not mobile_device.is_active:
            return mobile_device.change_and_save(**defaults)
        elif is_created_mobile_device:
            return mobile_device
        else:
            raise MobileDeviceAlreadyExists('Device key already exists')


class MobileDevice(SmartModel):
    """Model used to authenticate mobile devices. Unhashed login_token is stored
    in the device keychain and serve as password to log in together with UUID via DeviceBackend."""

    # this is not UUIDField because of the strict length limitation
    uuid = models.CharField(
        verbose_name=_('UUID'),
        max_length=36,
        null=False,
        blank=False
    )
    name = models.CharField(
        verbose_name=_('name'),
        max_length=250,
        null=True,
        blank=True
    )
    slug = models.SlugField(
        verbose_name=_('slug'),
        null=True,
        blank=True
    )
    last_login = models.DateTimeField(
        verbose_name=_('last login'),
        null=True,
        blank=True
    )
    user = models.ForeignKey(
        to=django_settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        verbose_name=_('user'),
        related_name='mobile_devices'
    )
    login_token = models.CharField(
        max_length=128,
        verbose_name=_('login token'),
        null=False,
        blank=False
    )
    is_active = models.BooleanField(
        verbose_name=_('is active'),
        default=True
    )
    user_agent = models.CharField(
        verbose_name=_('user agent'),
        max_length=256,
        null=True,
        blank=True,
    )
    is_primary = models.BooleanField(
        verbose_name=_('is primary'),
        default=False
    )

    @transaction.atomic()
    def clean_is_primary(self):
        if self.is_primary and 'is_primary' in self.changed_fields:
            # only one device can be primary
            self.__class__.objects.filter(user=self.user).exclude(pk=self.pk).select_for_update().change_and_save(
                is_primary=False
            )

    class Meta:
        unique_together = ('uuid', 'user')
        verbose_name = _('mobile device')
        verbose_name_plural = _('mobile devices')

    objects = MobileDeviceQuerySet.as_manager()

    def check_password(self, token):
        return check_password(str(token), self.login_token)


class OneTimePasswordManager(BaseHashKeyManager):

    def _hash_key(self, key, slug, **kwargs):
        return hash_key(key, salt=slug)

    def create(self, slug, key_generator, **kwargs):
        return super().create(key_generator, slug=slug, **kwargs)


class OneTimePassword(SmartModel):
    """
    Specific verification tokens that can be send via e-mail, SMS or another transmission medium
    to check user authorization (example password reset)
    """

    key = models.CharField(
        verbose_name=_('key'),
        max_length=128,
        unique=True,
        db_index=True,
        null=False,
        blank=False
    )
    expires_at = models.DateTimeField(
        verbose_name=_('expires at'),
        null=True,
        blank=True,
    )
    slug = models.SlugField(
        verbose_name=_('slug'),
        null=False,
        blank=False
    )
    is_active = models.BooleanField(
        verbose_name=_('is active'),
        default=True
    )
    data = models.JSONField(
        verbose_name=_('data'),
        null=True,
        blank=True,
        encoder=DjangoJSONEncoder
    )
    related_objects = GenericManyToManyField()

    secret_key = None

    objects = OneTimePasswordManager()

    class Meta:
        ordering = ('-created_at',)
        verbose_name = _('one time password')
        verbose_name_plural = _('one time passwords')

    @property
    def is_expired(self):
        return self.expires_at and self.expires_at < timezone.now()


class AuthorizationRequest(SmartModel):

    authorization_token = models.ForeignKey(
        verbose_name=_('authorization token'),
        to=AuthorizationToken,
        related_name='authorization_requests',
        null=True,
        blank=True,
        on_delete=models.SET_NULL
    )
    user = models.ForeignKey(
        verbose_name=_('user'),
        to=django_settings.AUTH_USER_MODEL,
        related_name='authorization_requests',
        null=False,
        blank=False,
        on_delete=models.CASCADE
    )
    slug = models.SlugField(
        verbose_name=_('slug'),
        null=True,
        blank=True
    )
    backend = models.CharField(
        verbose_name=_('backend'),
        max_length=250,
        null=False,
        blank=False
    )
    title = models.CharField(
        verbose_name=_('title'),
        max_length=250,
        null=False,
        blank=False
    )
    description = models.TextField(
        verbose_name=_('description'),
        null=True,
        blank=True
    )
    result = NumEnumField(
        verbose_name=_('result'),
        enum=AuthorizationRequestResult,
        null=True,
        blank=True
    )
    data = models.JSONField(
        verbose_name=_('data'),
        null=True,
        blank=True,
        encoder=DjangoJSONEncoder
    )
    expires_at = models.DateTimeField(
        verbose_name=_('expires at'),
        null=True,
        blank=True,
    )
    granted_at = models.DateTimeField(
        verbose_name=_('granted at'),
        null=True,
        blank=True
    )
    related_objects = GenericManyToManyField()

    class Meta:
        ordering = ('-created_at',)
        verbose_name = _('authorization request')
        verbose_name_plural = _('authorization requests')

    @property
    def is_expired(self):
        return self.expires_at < timezone.now()

    @property
    def state(self):
        if not self.result and self.is_expired:
            return AuthorizationRequestState.EXPIRED
        elif not self.result:
            return AuthorizationRequestState.WAITING
        else:
            return AuthorizationRequestState[self.result.name]



import binascii
import random
import re
import string

from datetime import timedelta
from functools import reduce

from operator import and_ as AND

import import_string

from django.core.exceptions import ImproperlyConfigured
from django.conf import settings as django_settings
from django.contrib.auth import _get_backends, load_backend
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth.signals import user_logged_in, user_logged_out
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.middleware.csrf import rotate_token
from django.utils.crypto import salted_hmac
from django.utils.timezone import now

from .config import settings
from .enums import AuthorizationRequestState, AuthorizationRequestResult
from .signals import authorization_denied, authorization_granted, authorization_cancelled

from ipware.ip import get_client_ip


def _load_request_backend(path):
    return import_string(path)


def get_authorization_request_backends():
    return {
        backend_path: _load_request_backend(backend_path)
        for backend_path in settings.AUTHORIZATION_REQUEST_BACKENDS
    }


def get_authorization_backend(path, raise_exception=False):
    backends = get_authorization_request_backends()

    backend = backends.get(path, None)
    if not backend and raise_exception:
        raise ImproperlyConfigured('Missing authorization request backend with path {}'.format(path))
    return backend() if backend else None


def compute_expires_at(expiration):
    """
    Compute expires at datetime from expiration in seconds.
    Args:
        expiration: expiration time in seconds.

    Returns:
        datetime value
    """
    return now() + timedelta(seconds=expiration)


def hash_key(key, salt=None):
    """
    Hash input key with HMAC. It is used constant SALT defined in configuration.
    It is used algorithm defined in Django DEFAULT_HASHING_ALGORITHM setting
    Django SECRET_KEY is used as a secret.
    Args:
        key: value to hash
        salt: salt value

    Returns:
        hashed value
    """
    return salted_hmac(
        settings.HASH_SALT,
        key,
        secret='|'.join((salt or '', django_settings.SECRET_KEY)),
        algorithm=django_settings.DEFAULT_HASHING_ALGORITHM,
    ).hexdigest()


def header_name_to_django(header_name):
    """
    Convert header into Django format with HTTP prefix.
    Args:
        header_name: HTTP header format (ex. X-Authorization)

    Returns:
        django header format (ex. HTTP_X_AUTHORIZATION)
    """
    return '_'.join(('HTTP', header_name.replace('-', '_').upper()))


def generate_key(characters=string.ascii_letters + string.digits, length=20):
    """
    Generate random string with given characters and length.
    Args:
        characters: characters used for key generator.
        length: key length.

    Returns:
        generated random key.
    """
    return ''.join(random.choices(characters, k=length))


def generate_otp_key():
    """
    Default OTP key generator.
    """
    return generate_key(
        characters=settings.OTP_DEFAULT_GENERATOR_CHARACTERS, length=settings.OTP_DEFAULT_GENERATOR_LENGTH
    )


def login(request, user, auth_slug=None, related_objs=None, backend=None, allowed_cookie=True,
          allowed_header=True, two_factor_login=False, expiration=None, preserve_cookie=False,
          mobile_device=None):
    """
    Persist token into database. Token is stored inside cookie therefore is not necessary
    reauthenticate user for every request.
    Args:
        request: Django HTTP request instance
        user: user to login.
        auth_slug: slug stored with authorization token.
        related_objs: django model instances shored with authorization token.
        backend: backend which was used for authorization.
        allowed_cookie: token can be used with the cookies.
        allowed_header: token can be used with the HTTP headers.
        two_factor_login: token will require second factor.
        expiration: expiration time in seconds.
        preserve_cookie: keep cookie after browser closing.
        mobile_device: mobile device which was used for authorization.

    Returns:
        new authorization token which contains secret_key.
    """
    from auth_token.models import AuthorizationToken, compute_authorization_token_expires_at

    related_objs = related_objs if related_objs is not None else ()

    if user is None:
        user = request.user

    try:
        backend = backend or user.backend
    except AttributeError:
        backends = _get_backends(return_tuples=True)
        if len(backends) == 1:
            _, backend = backends[0]
        else:
            raise ValueError(
                'You have multiple authentication backends configured and '
                'therefore must provide the `backend` argument or set the '
                '`backend` attribute on the user.'
            )

    token = AuthorizationToken.objects.create(
        user=user,
        user_agent=request.META.get('HTTP_USER_AGENT', '')[:256],
        auth_slug=auth_slug,
        ip=get_client_ip(request)[0],
        backend=backend,
        allowed_cookie=allowed_cookie,
        allowed_header=allowed_header,
        is_authenticated=not two_factor_login,
        expires_at=compute_authorization_token_expires_at(expiration),
        preserve_cookie=preserve_cookie,
        mobile_device=mobile_device or getattr(user, 'authenticated_mobile_device', None)
    )

    if related_objs:
        token.related_objects.add(*related_objs)
    if token.is_authenticated:
        if hasattr(request, 'user'):
            request.user = user
        user_logged_in.send(sender=user.__class__, request=request, user=user)
    request.token = token
    rotate_token(request)


def authorize_login(authorization_token, request=None):
    """
    Set authorization token to be fully authenticated.
    Args:
        request: django HTTP request.
        authorization_token: authorization token to be authenticated
    """
    if not authorization_token.is_authenticated:
        authorization_token.change_and_save(is_authenticated=True, update_only_changed_fields=True)
        user_logged_in.send(sender=authorization_token.user.__class__, request=request, user=authorization_token.user)
        if request and authorization_token == request.token:
            request.token = authorization_token
            request.user = get_user_from_token(authorization_token)


def logout(request):
    """
    Set current token as an inactive.
    Args:
        request: Django HTTP request instance.
    """
    from auth_token.models import AnonymousAuthorizationToken

    # Dispatch the signal before the user is logged out so the receivers have a
    # chance to find out *who* logged out.
    user = getattr(request, 'user', None)
    if hasattr(user, 'is_authenticated') and not user.is_authenticated:
        user = None
    user_logged_out.send(sender=user.__class__, request=request, user=user)

    if hasattr(request, 'token') and request.token.is_active:
        if request.token.active_takeover:
            request.token.active_takeover.change_and_save(is_active=False)
            if hasattr(request, 'user'):
                request.user = request.token.user
        else:
            token = request.token
            token.change_and_save(is_active=False)
            request.token = AnonymousAuthorizationToken()

    if hasattr(request, 'user'):
        request.user = AnonymousUser()


def create_auth_header_value(token):
    """
    Create value for HTTP header.
    Args:
        token: authorization token key. (ex. 123456)

    Returns:
        HTTP header value (ex. Bearer 123456)
    """
    return token if settings.HEADER_TOKEN_TYPE is None else '{} {}'.format(settings.HEADER_TOKEN_TYPE, token)


def parse_auth_header_value(request):
    """
    Parse a token parsed from the "Authorization" header.
    Args:
        request: Django HTTP request instance.

    Returns:
        authorization token key.
    """
    header_value = request.META.get(header_name_to_django(settings.HEADER_NAME))

    if not header_value:
        raise ValueError('Authorization header missing')

    if settings.HEADER_TOKEN_TYPE is None:
        return header_value
    else:
        match = re.match(
            '{} ([^ ]+)$'.format(settings.HEADER_TOKEN_TYPE),
            request.META.get(header_name_to_django(settings.HEADER_NAME), '')
        )
    return match.group(1) if match else None


def get_token_key_from_request(request):
    """
    Return authorization token key from Django request instance.
    Args:
        request: Django HTTP request instance.

    Returns:
        tuple with values (token key, boolean if request was get from header, boolean if request was get from cookie)
    """
    if settings.HEADER and header_name_to_django(settings.HEADER_NAME) in request.META:
        return parse_auth_header_value(request), True, False
    elif settings.COOKIE and settings.COOKIE_NAME in request.COOKIES:
        return request.COOKIES.get(settings.COOKIE_NAME), False, True
    else:
        return None, False, False


def get_token(request):
    """
    Returns the token model instance associated with the given request token key.
    If no user is retrieved AnonymousAuthorizationToken is returned.
    Args:
        request: Django HTTP request instance.

    Returns:
        Instance of authorization token.
    """
    from auth_token.models import AnonymousAuthorizationToken, AuthorizationToken, hash_key
    auth_token, token_is_from_header, token_is_from_cookie = get_token_key_from_request(request)

    try:
        token = AuthorizationToken.objects.get(
            key=hash_key(auth_token),
            is_active=True,
            allowed_cookie__gte=token_is_from_cookie,
            allowed_header__gte=token_is_from_header,
            expires_at__gt=now()
        )
        token.is_from_header, token.is_from_cookie = token_is_from_header, token_is_from_cookie
        token.secret_key = auth_token
        return token
    except AuthorizationToken.DoesNotExist:
        return AnonymousAuthorizationToken()


def dont_enforce_csrf_checks(request):
    """
    Check if CSRF validation is required. If token is get from HTTP header CSRF check is not necessary.
    Args:
        request: Django HTTP request instance.

    Returns:
        boolean if csrf check should be processed or none.
    """
    return (
        header_name_to_django(settings.HEADER_NAME) in request.META or
        getattr(request, '_dont_enforce_csrf_checks', False)
    )


def get_user_from_token(token):
    """
    Return user from authorization token.
    Args:
        token: authorization token instance.

    Returns:
        user instance or AnonymousUser.
    """
    if token:
        backend_path = token.backend
        if backend_path in django_settings.AUTHENTICATION_BACKENDS and token.is_authenticated:
            active_takeover_id = token.active_takeover.user.pk if token.active_takeover else None
            user_id = token.user.pk
            backend = load_backend(backend_path)
            return backend.get_user(active_takeover_id) or backend.get_user(user_id) or AnonymousUser()
    return AnonymousUser()


def get_user(request):
    """
    Returns the user model instance associated with the given request token.
    If no user is retrieved an instance of `AnonymousUser` is returned.
    Args:
        request: Django HTTP request instance.

    Returns:
        user instance or AnonymousUser.
    """
    return get_user_from_token(getattr(request, 'token'))


def takeover(request, user):
    """
    Login the current user to any other user.
    Args:
        request: Django HTTP request instance with already authorized user.
        user: takeovered user.

    Returns:
        True if takeover was performed or False.
    """
    if not hasattr(request, 'user') or not request.user.is_authenticated or request.user == user:
        return False
    else:
        request.token.user_takeovers.update(is_active=False)
        request.token.user_takeovers.create(user=user, is_active=True)
        return True


def deactivate_otp(slug, key=None, related_objects=None):
    """
    Deactivate one time password with a given slug, key and related objects.
    Args:
        slug: OTP slug
        key: OTP key or None.
        related_objects: list of related objects which must be related with OTP to be deactivated.
    """
    from .models import OneTimePassword

    otp_qs = OneTimePassword.objects.filter(slug=slug, is_active=True)

    if related_objects is not None:
        otp_qs = otp_qs.filter(
            reduce(AND, (
                Q(**{
                    'related_objects__object_id': obj.pk,
                    'related_objects__object_ct': ContentType.objects.get_for_model(obj)
                }) for obj in related_objects
            ))
        )
    if key is not None:
        otp_qs = otp_qs.filter(key=hash_key(key, salt=slug))

    otp_qs.change_and_save(is_active=False)


def create_otp(slug, related_objects=None, data=None, key_generator=None, expiration=None, deactivate_old=False):
    """
    Create new one time password. One time password must be identified with slug.
    Args:
        slug: string for OTP identification.
        related_objects: model instances related with OTP.
        data: data which will be stored with OTP in the JSON format.
        key_generator: OTP key generator.
        expiration: OTP expiration time in seconds, default expiration will be used for None value.
        deactivate_old: deactivate old tokens with the same slug ane related objects.

    Returns:
        OTP instance
    """
    from .models import OneTimePassword, compute_expires_at

    if deactivate_old:
        deactivate_otp(slug, related_objects=related_objects)

    key_generator = import_string(settings.OTP_DEFAULT_KEY_GENERATOR) if key_generator is None else key_generator

    otp = OneTimePassword.objects.create(
        slug=slug,
        key_generator=key_generator,
        expires_at=compute_expires_at(expiration or settings.OTP_DEFAULT_AGE),
        data=data
    )
    if related_objects:
        otp.related_objects.add(*related_objects)
    return otp


def get_otp_qs(slug=None, key=None, related_objects=None, **kwargs):
    """
    Return OTP queryset with slug, key or related objects
    Args:
        slug: string for OTP identification.
        key: key used for validation of OTP.
        related_objects: objects related with OTP.
        **kwargs: extra filter kwargs.

    Returns:
        OneTimePassword model queryset
    """
    from .models import OneTimePassword

    otp_qs = OneTimePassword.objects.filter(**kwargs)

    if slug:
        otp_qs = otp_qs.filter(slug=slug)

    if key:
        otp_qs = otp_qs.filter(key=hash_key(key, salt=slug))

    if related_objects is not None:
        otp_qs = otp_qs.filter(
            reduce(AND, (
                Q(**{
                    'related_objects__object_id': obj.pk,
                    'related_objects__object_ct': ContentType.objects.get_for_model(obj)
                }) for obj in related_objects
            ))
        )
    return otp_qs


def get_valid_otp(slug, key=None, related_objects=None, deactivate=False):
    """
    Find and return valid OTP with a given slug and key.
    Args:
        slug: string for OTP identification.
        key: key used for validation of OTP.
        related_objects: list of related objects which must be related with OTP.
        deactivate: deactivate the OTP after successful retrieval.

    Returns:
        valid OTP instance or None value.
    """
    otp =  get_otp_qs(
        slug=slug,
        key=key,
        related_objects=related_objects,
        is_active=True,
        expires_at__gt=now()
    ).first('-created_at')

    if otp and deactivate:
        otp.change_and_save(is_active=False)

    return otp


def check_otp(slug, key):
    """
    Check if key is valid for OTP.
    Args:
        slug: string for OTP identification.
        key: key used for validation of OTP.

    Returns:
        True for valid OTP code with a given key, False otherwise.
    """
    return get_valid_otp(slug, key) is not None


def extend_otp(slug, key, expiration=None):
    """
    Extend OTP key with a given expiration time in seconds.
    Args:
        slug: string for OTP identification.
        key: key used for validation of OTP.
        expiration: OTP expiration time in seconds, default expiration will be used for None value.

    Returns:
        True if OTP expiration was extended, False otherwise
    """
    otp = get_valid_otp(slug, key)
    if otp is None:
        return False
    else:
        otp.change_and_save(expires_at=compute_expires_at(expiration or settings.DEFAULT_OTP_AGE))
        return True


def create_authorization_request(slug, user, title, description=None, authorization_token=None,
                                 related_objects=None, data=None, expiration=None, backend_path=None):
    """
    Create authorization request with a given slug, title and description.
    Args:
        slug: string for authorization request identification.
        user: owner of authorization request.
        title: human readable title for a client.
        description: human readable description for a client.
        authorization_token: authorization request can be valid only for specific authorization token.
        related_objects: related model instances which will be stored with authorization request.
        data: extra authorization request data.
        expiration: expiration time in seconds.
        backend_path: path authorization request backend class with authentication implementation.

    Returns:
        new authorization request.
    """
    from .models import AuthorizationRequest

    if not backend_path:
        if len(settings.AUTHORIZATION_REQUEST_BACKENDS) == 1:
            backend_path = settings.AUTHORIZATION_REQUEST_BACKENDS[0]
        else:
            raise ImproperlyConfigured(
                'You must select backend_path if more than one AUTH_TOKEN_AUTHORIZATION_REQUEST_BACKENDS are set'
            )

    backend = get_authorization_backend(backend_path, raise_exception=True)
    expiration = expiration or settings.DEFAULT_AUTHORIZATION_REQUEST_AGE

    authorization_request = AuthorizationRequest.objects.create(
        slug=slug,
        user=user,
        title=title,
        description=description,
        authorization_token=authorization_token,
        data=data,
        expires_at=compute_expires_at(expiration),
        backend=backend_path
    )
    if related_objects:
        authorization_request.related_objects.add(*related_objects)

    backend.initialize(authorization_request)
    return authorization_request


def reset_authorization_request(authorization_request, expiration=None):
    """
    Reset authorization request and increase its expiration.
    Args:
        authorization_request: authorization request where new OTP will be created.
        expiration: expiration time in seconds.

    Returns:
        AuthorizationRequest instance with new authorization data
    """
    assert authorization_request.result is None
    backend = get_authorization_backend(authorization_request.backend, raise_exception=True)
    expiration = expiration or settings.DEFAULT_AUTHORIZATION_REQUEST_AGE

    authorization_request.change_and_save(expires_at=compute_expires_at(expiration))
    backend.initialize(authorization_request)
    return authorization_request


def check_authorization_request(authorization_request, **kwargs):
    """
    Check if data send by client is valid with autorization request and access can be granted.
    Args:
        authorization_request: validated authorization request
        **kwargs: data used for request authentication.

    Returns:
        True for valid input data, False otherwise.
    """
    if authorization_request.state != AuthorizationRequestState.WAITING:
        return False

    backend = get_authorization_backend(authorization_request.backend, raise_exception=False)

    if not backend:
        return False
    return backend.authenticate(authorization_request, **kwargs)


def grant_authorization_request(authorization_request, **kwargs):
    """
    Grant access for the authorization request.
    Args:
        authorization_request: authorization request to grant.
        **kwargs: extra data which will be send to the signal.
    """
    assert authorization_request.result is None

    backend = get_authorization_backend(authorization_request.backend, raise_exception=False)
    if backend:
        backend.grant(authorization_request)
    authorization_request.change_and_save(
        result=AuthorizationRequestResult.GRANTED,
        granted_at=now()
    )
    authorization_granted.send(sender=authorization_request.slug, authorization_request=authorization_request, **kwargs)


def deny_authorization_request(authorization_request, **kwargs):
    """
    Deny access for the authorization request.
    Args:
       authorization_request: authorization request to deny.
       **kwargs: extra data which will be send to the signal.
    """
    assert authorization_request.result is None

    backend = get_authorization_backend(authorization_request.backend, raise_exception=False)
    if backend:
        backend.deny(authorization_request)
    authorization_request.change_and_save(result=AuthorizationRequestResult.DENIED)
    authorization_denied.send(sender=authorization_request.slug, authorization_request=authorization_request)


def cancel_authorization_request(authorization_request, **kwargs):
    """
    Cancel authorization request.
    Args:
        authorization_request: authorization request to cancel.
        **kwargs: extra data which will be send to the signal.
    """
    assert authorization_request.result is None

    backend = get_authorization_backend(authorization_request.backend, raise_exception=False)
    if backend:
        backend.cancel(authorization_request)
    authorization_request.change_and_save(result=AuthorizationRequestResult.CANCELLED)
    authorization_cancelled.send(sender=authorization_request.slug, authorization_request=authorization_request)



# Generated by Django 3.1 on 2021-01-19 14:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('auth_token', '0012_auto_20210119_1308'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='tokenrelatedobject',
            name='content_type',
        ),
        migrations.RemoveField(
            model_name='tokenrelatedobject',
            name='token',
        ),
        migrations.RemoveField(
            model_name='usertokentakeover',
            name='token',
        ),
        migrations.RemoveField(
            model_name='usertokentakeover',
            name='user',
        ),
        migrations.AlterModelOptions(
            name='mobiledevice',
            options={'verbose_name': 'mobile device', 'verbose_name_plural': 'mobile devices'},
        ),
        migrations.AlterField(
            model_name='mobiledevice',
            name='changed_at',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at'),
        ),
        migrations.AlterField(
            model_name='mobiledevice',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at'),
        ),
        migrations.DeleteModel(
            name='Token',
        ),
        migrations.DeleteModel(
            name='TokenRelatedObject',
        ),
        migrations.DeleteModel(
            name='UserTokenTakeover',
        ),
        migrations.AlterField(
            model_name='authorizationtoken',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at'),
        ),
        migrations.AlterField(
            model_name='authorizationtoken',
            name='changed_at',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at'),
        ),
    ]



from sys import intern

from collections import defaultdict

from django.dispatch.dispatcher import Signal


class StringSenderSignal(Signal):

    _used_senders = defaultdict(object)

    def _get_unique_sender(self, sender):
        if sender is None:
            return None
        return self._used_senders[sender]

    def connect(self, receiver, sender=None, weak=True, dispatch_uid=None):
        super().connect(receiver, self._get_unique_sender(sender), weak, dispatch_uid)

    def disconnect(self, receiver=None, sender=None, dispatch_uid=None):
        return super().disconnect(receiver, self._get_unique_sender(sender), dispatch_uid)

    def send(self, sender, **named):
        return super().send(self._get_unique_sender(sender), **named)


authorization_granted = StringSenderSignal()
authorization_denied = StringSenderSignal()
authorization_cancelled = StringSenderSignal()


# Generated by Django 3.1 on 2021-03-08 08:18

from django.db import migrations, models
from django.conf import settings
import auth_token.enums
import auth_token.models
import django.db.models.deletion
import enumfields.fields


class Migration(migrations.Migration):

    dependencies = [
        ('auth_token', '0013_auto_20210119_1547'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='authorizationrequestgenericmanytomanyrelation',
            unique_together=None,
        ),
        migrations.RemoveField(
            model_name='authorizationrequestgenericmanytomanyrelation',
            name='authorization_request',
        ),
        migrations.RemoveField(
            model_name='authorizationrequestgenericmanytomanyrelation',
            name='object_ct',
        ),
        migrations.AlterUniqueTogether(
            name='onetimepasswordgenericmanytomanyrelation',
            unique_together=None,
        ),
        migrations.RemoveField(
            model_name='onetimepasswordgenericmanytomanyrelation',
            name='object_ct',
        ),
        migrations.RemoveField(
            model_name='onetimepasswordgenericmanytomanyrelation',
            name='one_time_password',
        ),
        migrations.AlterUniqueTogether(
            name='authorizationtokengenericmanytomanyrelation',
            unique_together=None,
        ),
        migrations.RemoveField(
            model_name='authorizationtokengenericmanytomanyrelation',
            name='object_ct',
        ),
        migrations.RemoveField(
            model_name='authorizationtokengenericmanytomanyrelation',
            name='authorization_token',
        ),
        migrations.AddField(
            model_name='authorizationtoken',
            name='id',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.DeleteModel(
            name='AuthorizationRequest',
        ),
        migrations.DeleteModel(
            name='AuthorizationRequestGenericManyToManyRelation',
        ),
        migrations.DeleteModel(
            name='OneTimePassword',
        ),
        migrations.DeleteModel(
            name='OneTimePasswordGenericManyToManyRelation',
        ),
        migrations.DeleteModel(
            name='AuthorizationTokenGenericManyToManyRelation',
        ),
        migrations.DeleteModel(
            name='UserAuthorizationTokenTakeover',
        ),
        migrations.DeleteModel(
            name='AuthorizationToken'
        )
    ]



from django.contrib.auth.backends import ModelBackend
from django.core.exceptions import PermissionDenied
from django.utils import timezone

from auth_token.models import MobileDevice


class DeviceBackend(ModelBackend):
    """
    Authenticates device with UUID and login_token (password)
    """

    def authenticate(self, request, mobile_device_uuid=None, mobile_login_token=None, **kwargs):
        if not mobile_device_uuid or not mobile_login_token:
            return None

        mobile_device_qs = MobileDevice.objects.filter(uuid=mobile_device_uuid, is_active=True)
        if not mobile_device_qs.exists():
            raise PermissionDenied('MobileDevice with id "{}" not found.'.format(mobile_device_uuid))

        for mobile_device in mobile_device_qs:
            if mobile_device.check_password(mobile_login_token):
                user = self.get_user(mobile_device.user_id)
                if user:
                    mobile_device.change_and_save(
                        last_login=timezone.now(),
                        update_only_changed_fields=True
                    )
                    user.authenticated_mobile_device = mobile_device
                    return user
                else:
                    return None

        raise PermissionDenied('Provided invalid login_token to MobileDevice with id "{}".'.format(mobile_device_uuid))



from django import forms
from django.contrib.auth.hashers import check_password
from django.utils.translation import ugettext as _

from is_core.forms.forms import SmartForm

from auth_token.config import settings
from auth_token.contrib.common.forms import AuthenticationCleanMixin, TokenAuthenticationMixin
from auth_token.models import AuthorizationToken, hash_key
from auth_token.utils import authorize_login, get_valid_otp, check_authorization_request


class TokenAuthenticationSmartForm(TokenAuthenticationMixin, AuthenticationCleanMixin, SmartForm):
    pass


class LoginCodeVerificationForm(SmartForm):

    code = forms.CharField(max_length=40, required=True, label=_('code'))

    def __init__(self, request, *args, **kwargs):
        self.request = request
        super().__init__(*args, **kwargs)

    def _init_code(self, field):
        if settings.AUTHORIZATION_REQUEST_OTP_DEBUG_CODE:
            field.initial = settings.AUTHORIZATION_REQUEST_OTP_DEBUG_CODE

    def get_authorization_request(self):
        return self.request.token.authorization_requests.filter(
            slug=settings.TWO_FACTOR_AUTHORIZATION_SLUG
        ).first('-created_at')

    def get_user(self):
        return self.request.token.user

    def clean_code(self):
        code = self.cleaned_data.get('code')
        authorization_requests = self.get_authorization_request()

        if authorization_requests and check_authorization_request(authorization_requests, otp_secret_key=code):
            authorize_login(self.request.token, self.request)
        else:
            raise forms.ValidationError(_('The inserted value does not correspond to the sent code.'))
        return code



# Generated by Django 3.1 on 2021-03-08 11:43

import auth_token.enums
import auth_token.models
from django.conf import settings
import django.core.serializers.json
from django.db import migrations, models
import django.db.models.deletion
import enumfields.fields


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
        ('auth_token', '0014_auto_20210308_0918'),
    ]

    operations = [
        migrations.CreateModel(
            name='AuthorizationToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('key', models.CharField(db_index=True, max_length=128, unique=True, verbose_name='key')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('user_agent', models.CharField(blank=True, max_length=256, null=True, verbose_name='user agent')),
                ('expires_at', models.DateTimeField(default=auth_token.models.compute_authorization_token_expires_at,
                                                    verbose_name='expires at')),
                ('ip', models.GenericIPAddressField(verbose_name='IP')),
                ('auth_slug', models.SlugField(blank=True, null=True, verbose_name='slug')),
                ('backend', models.CharField(max_length=250, verbose_name='backend')),
                ('allowed_cookie', models.BooleanField(default=True, verbose_name='is allowed cookie')),
                ('allowed_header', models.BooleanField(default=True, verbose_name='is allowed header')),
                ('is_authenticated', models.BooleanField(default=False, verbose_name='is authenticated')),
                ('preserve_cookie', models.BooleanField(default=False, verbose_name='preserve cookie')),
                ('mobile_device', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE,
                                                    related_name='authorization_tokens', to='auth_token.mobiledevice',
                                                    verbose_name='mobile device')),
                ('user',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='authorization_tokens',
                                   to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'verbose_name': 'authorization token',
                'verbose_name_plural': 'authorization tokens',
            },
        ),
        migrations.CreateModel(
            name='OneTimePassword',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('key', models.CharField(db_index=True, max_length=128, unique=True, verbose_name='key')),
                ('expires_at', models.DateTimeField(blank=True, null=True, verbose_name='expires at')),
                ('slug', models.SlugField(verbose_name='slug')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('data', models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                          verbose_name='data')),
            ],
            options={
                'verbose_name': 'one time password',
                'verbose_name_plural': 'one time passwords',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='UserAuthorizationTokenTakeover',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('is_active', models.BooleanField()),
                ('token', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='user_takeovers',
                                            to='auth_token.authorizationtoken', verbose_name='authorization token')),
                ('user',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='user_token_takeovers',
                                   to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'verbose_name': 'authorization takeover',
                'verbose_name_plural': 'authorization takeovers',
            },
        ),
        migrations.CreateModel(
            name='AuthorizationRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('slug', models.SlugField(blank=True, null=True, verbose_name='slug')),
                ('backend', models.CharField(max_length=250, verbose_name='backend')),
                ('title', models.CharField(max_length=250, verbose_name='title')),
                ('description', models.TextField(blank=True, null=True, verbose_name='description')),
                ('result',
                 enumfields.fields.NumEnumField(blank=True, enum=auth_token.enums.AuthorizationRequestResult, null=True,
                                                verbose_name='result')),
                ('data', models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                          verbose_name='data')),
                ('expires_at', models.DateTimeField(blank=True, null=True, verbose_name='expires at')),
                ('granted_at', models.DateTimeField(blank=True, null=True, verbose_name='granted at')),
                ('authorization_token',
                 models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                   related_name='authorization_requests', to='auth_token.authorizationtoken',
                                   verbose_name='authorization token')),
                ('user',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='authorization_requests',
                                   to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'verbose_name': 'authorization request',
                'verbose_name_plural': 'authorization requests',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='OneTimePasswordGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('object_ct',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype',
                                   verbose_name='content type of the related object')),
                ('one_time_password',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='auth_token.onetimepassword')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('one_time_password', 'object_ct', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='AuthorizationTokenGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('authorization_token',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='auth_token.authorizationtoken')),
                ('object_ct',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype',
                                   verbose_name='content type of the related object')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('authorization_token', 'object_ct', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='AuthorizationRequestGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('authorization_request',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='auth_token.authorizationrequest')),
                ('object_ct',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype',
                                   verbose_name='content type of the related object')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('authorization_request', 'object_ct', 'object_id')},
            },
        ),
    ]



from datetime import timedelta

from django.utils import timezone
from django.core.management.base import BaseCommand

from auth_token.config import settings
from auth_token.models import AuthorizationRequest


class Command(BaseCommand):

    def handle(self, **options):
        authorization_request_to_delete_qs = AuthorizationRequest.objects.filter(
            expires_at__lt=timezone.now() - timedelta(seconds=settings.AUTHORIZATION_REQUEST_PRESERVE_AGE)
        )
        self.stdout.write('Will delete {} authorization requests'.format(
            authorization_request_to_delete_qs.count())
        )
        deletion_count = authorization_request_to_delete_qs.delete()
        self.stdout.write('Deleted {} authorization requests'.format(deletion_count[0]))
        self.stdout.write('{} authorization requests remain in database'.format(AuthorizationRequest.objects.count()))



from auth_token.config import settings
from auth_token.utils import create_otp, deactivate_otp, get_valid_otp
from auth_token.models import MobileDevice


class BaseAuthorizationRequestBackend:

    def initialize(self, authorization_request):
        """
        Initialize backend for the authorization request. It is called with create new autorization or its reset.
        Args:
            authorization_request: AuthorizationRequest instance.
        """
        pass

    def authenticate(self, authorization_request, **kwargs):
        """
        Check if authorization input data is valid and can grant its access.
        Args:
            authorization_request: AuthorizationRequest instance.
            **kwargs: input data to check access.

        Return:
            True/False if input data are valid.
        """
        raise NotImplementedError

    def grant(self, authorization_request):
        """
        Method is called when authorization request was granted.
        Args:
            authorization_request: AuthorizationRequest instance.
        """
        pass

    def deny(self, authorization_request):
        """
        Method is called when authorization request was denied.
        Args:
            authorization_request: AuthorizationRequest instance.
        """
        pass

    def cancel(self, authorization_request):
        """
        Method is called when authorization request was cancelled.
        Args:
            authorization_request: AuthorizationRequest instance.
        """
        pass


class OTPAuthorizationRequestBackend(BaseAuthorizationRequestBackend):
    """
    Backend for authorization request authentication via OTP.
    """

    def _create_otp(self, authorization_request):
        return create_otp(
            authorization_request.slug,
            expiration=(authorization_request.expires_at - authorization_request.created_at).total_seconds(),
            related_objects=[authorization_request],
            deactivate_old=True
        )

    def initialize(self, authorization_request):
        otp = self._create_otp(authorization_request)
        authorization_request.secret_key = otp.secret_key

    def authenticate(self, authorization_request, otp_secret_key=None, **kwargs):
        if otp_secret_key is None:
            return False

        if (settings.AUTHORIZATION_REQUEST_OTP_DEBUG_CODE
                and settings.AUTHORIZATION_REQUEST_OTP_DEBUG_CODE == otp_secret_key):
            otp_secret_key = None

        otp = get_valid_otp(authorization_request.slug, key=otp_secret_key, related_objects=[authorization_request])
        return otp is not None

    def _deactivate(self, authorization_request):
        deactivate_otp(authorization_request.slug, related_objects=[authorization_request])

    def grant(self, authorization_request):
        self._deactivate(authorization_request)

    def deny(self, authorization_request):
        self._deactivate(authorization_request)

    def cancel(self, authorization_request):
        self._deactivate(authorization_request)


class MobileDeviceAuthorizationRequestBackend(BaseAuthorizationRequestBackend):
    """
    Backend for authorization request authentication vie user mobile device.
    """

    def _get_valid_mobile_device(self, mobile_device_uuid=None, mobile_login_token=None):
        for mobile_device in MobileDevice.objects.filter(uuid=mobile_device_uuid, is_active=True):
            if mobile_device.check_password(mobile_login_token):
                return mobile_device
        return None

    def authenticate(self, authorization_request, mobile_device_uuid=None, mobile_login_token=None, **kwargs):
        valid_mobile_device = None
        if mobile_device_uuid and mobile_login_token:
            valid_mobile_device = self._get_valid_mobile_device(mobile_device_uuid, mobile_login_token)

        return valid_mobile_device is not None and valid_mobile_device.user == authorization_request.user



from setuptools import setup, find_packages

from auth_token.version import get_version

setup(
    name='django-auth-token',
    version=get_version(),
    description="Django authorization via tokens.",
    keywords='django, authorization',
    author='Lubos Matl',
    author_email='matllubos@gmail.com',
    url='https://github.com/druids/django-auth-token',
    license='BSD',
    package_dir={'auth_token': 'auth_token'},
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
    ],
    install_requires=[
        'django>=2.2.14',
        'django-ipware>=3.0.2',
        'import_string==0.1.0',
        'django-chamber>=0.6.5',
        'django-generic-m2m-field>=0.0.4',
        'django-choice-enumfields>=1.0.3',
    ],
    zip_safe=False
)



from datetime import timedelta

from django.db.models import Q
from django.utils import timezone
from django.core.management.base import BaseCommand

from auth_token.config import settings
from auth_token.models import OneTimePassword


class Command(BaseCommand):

    def handle(self, **options):
        inactive_otp_qs = OneTimePassword.objects.filter(
            Q(is_active=False)
            | Q(expires_at__isnull=False, expires_at__lt=timezone.now() - timedelta(
                seconds=settings.OTP_EXPIRATION_RETENTION_PERIOD,
            ))
        )
        self.stdout.write('Will delete {} inactive or expired OTP'.format(
            inactive_otp_qs.count())
        )
        deletion_count = inactive_otp_qs.delete()
        self.stdout.write('Deleted {} inactive or expired OTP'.format(deletion_count[0]))
        self.stdout.write('{} OTP remain in database'.format(OneTimePassword.objects.count()))



from django.utils.translation import ugettext_lazy as _

from enumfields import Choice, ChoiceEnum


class AuthorizationRequestResult(ChoiceEnum):

    GRANTED = Choice(1, _('granted'))
    DENIED = Choice(2, _('denied'))
    CANCELLED = Choice(3, _('cancelled'))


class AuthorizationRequestState(ChoiceEnum):

    GRANTED = Choice(1, _('granted'))
    DENIED = Choice(2, _('denied'))
    CANCELLED = Choice(3, _('cancelled'))
    WAITING = Choice(4, _('waiting'))
    EXPIRED = Choice(5, _('expired'))
